"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import base64
import logging
import traceback
import uuid

from datetime import datetime

from abc import ABCMeta, abstractmethod
from agora.stoa.actions.core import RDF, STOA, FOAF, TYPES, XSD
from agora.stoa.actions.core.base import Request, Action, Response, Sink
from agora.stoa.actions.core.utils import CGraph
from agora.stoa.messaging.reply import reply
from agora.stoa.server import app
from agora.stoa.store import r
from rdflib import BNode, Literal, RDFS

__author__ = 'Fernando Serena'

EXCHANGE_CONFIG = app.config['EXCHANGE']
exchange = EXCHANGE_CONFIG['exchange']
response_rk = EXCHANGE_CONFIG['response_rk']
log = logging.getLogger('agora.stoa.actions.delivery')
AGENT_UUID = Literal(str(uuid.uuid4()), datatype=TYPES.UUID)


def _build_reply_templates():
    """
    :return: Accept and Failure message templates
    """
    accepted = CGraph()
    failure = CGraph()
    response_node = BNode()
    agent_node = BNode()
    accepted.add((response_node, RDF.type, STOA.Root))
    accepted.add((response_node, RDF.type, STOA.Accepted))
    accepted.add((agent_node, RDF.type, FOAF.Agent))
    accepted.add((response_node, STOA.responseNumber, Literal("0", datatype=XSD.unsignedLong)))
    accepted.add((response_node, STOA.submittedBy, agent_node))
    accepted.add(
        (agent_node, STOA.agentId, AGENT_UUID))
    accepted.bind('types', TYPES)
    accepted.bind('stoa', STOA)
    accepted.bind('foaf', FOAF)

    for triple in accepted:
        failure.add(triple)
    failure.set((response_node, RDF.type, STOA.Failure))
    for (prefix, ns) in accepted.namespaces():
        failure.bind(prefix, ns)

    return accepted, failure


def build_reply(template, reply_to, comment=None):
    """
    :param template: A specific message template graph
    :param reply_to: Recipient Agent UUID
    :param comment: Optional comment
    :return: The reply graph
    """
    reply_graph = CGraph()
    root_node = None
    for (s, p, o) in template:
        if o == STOA.Root:
            root_node = s
        else:
            reply_graph.add((s, p, o))

    reply_graph.add((root_node, STOA.responseTo, Literal(reply_to, datatype=TYPES.UUID)))
    reply_graph.set((root_node, STOA.submittedOn, Literal(datetime.now())))
    reply_graph.set((root_node, STOA.messageId, Literal(str(uuid.uuid4()), datatype=TYPES.UUID)))
    if comment is not None:
        reply_graph.set((root_node, RDFS.comment, Literal(comment, datatype=XSD.string)))
    for (prefix, ns) in template.namespaces():
        reply_graph.bind(prefix, ns)
    return reply_graph


# Create both accept and failure templates
accepted_template, failure_template = _build_reply_templates()
log.info('Basic delivery templates created')


class DeliveryRequest(Request):
    def __init__(self):
        super(DeliveryRequest, self).__init__()

    def _extract_content(self, request_type=None):
        """
        Extracts delivery related data from message (delivery channel to reply to)
        """
        super(DeliveryRequest, self)._extract_content(request_type)

        q_res = self._graph.query("""SELECT ?node ?ex ?rk ?h ?p ?v WHERE {
                                  ?node stoa:replyTo [
                                        a stoa:DeliveryChannel;
                                        amqp:exchangeName ?ex;
                                        amqp:routingKey ?rk;
                                        amqp:broker [
                                           a amqp:Broker;
                                           amqp:host ?h;
                                           amqp:port ?p;
                                           amqp:virtualHost ?v
                                        ]
                                     ]
                                  } """)
        q_res = list(q_res)
        if len(q_res) != 1:
            raise SyntaxError('Invalid delivery request')

        request_fields = q_res.pop()

        if not any(request_fields):
            raise SyntaxError('Missing fields for delivery request')

        if request_fields[0] != self._request_node:
            raise SyntaxError('Request node does not match')

        delivery_data = {}

        (delivery_data['exchange'],
         delivery_data['routing_key'],
         delivery_data['host'],
         delivery_data['port'],
         delivery_data['vhost']) = request_fields[1:]
        log.debug("""Parsed attributes of a delivery action request:
                      -exchange name: {}
                      -routing key: {}
                      -host: {}
                      -port: {}
                      -virtual host: {}""".format(
            delivery_data['exchange'],
            delivery_data['routing_key'],
            delivery_data['host'], delivery_data['port'], delivery_data['vhost']))

        # Copy delivery data dictionary to the base request fields attribute
        self._fields['delivery'] = delivery_data.copy()

    @property
    def broker(self):
        """
        :return: Broker to which response must be addressed
        """
        broker_dict = {k: self._fields['delivery'][k].toPython() for k in ('host', 'port', 'vhost') if
                       k in self._fields['delivery']}
        broker_dict['port'] = int(broker_dict['port'])
        return broker_dict

    @property
    def channel(self):
        """
        :return: Delivery channel attributes
        """
        return {k: self._fields['delivery'][k].toPython() for k in ('exchange', 'routing_key') if
                k in self._fields['delivery']}

    @property
    def recipient(self):
        """
        :return: Broker and delivery channel data
        """
        recipient = self.broker.copy()
        recipient.update(self.channel)
        return recipient


class DeliveryAction(Action):
    __metaclass__ = ABCMeta

    def __init__(self, message):
        super(DeliveryAction, self).__init__(message)

    def __reply(self, template, reason=None):
        """
        Sends a protocol reply message to the submitter
        """
        graph = build_reply(template, self.request.message_id, reason)
        reply(graph.serialize(format='turtle'), exchange=exchange,
              routing_key='{}.{}'.format(response_rk, self.request.submitted_by),
              **self.request.broker)

    def __reply_accepted(self):
        """
        Sends an Accept message to the submitter
        """
        try:
            self.__reply(accepted_template)
        except Exception:
            # KeyError: Delivery channel data may be invalid
            # IOError: If the acceptance couldn't be sent, propagate the exception
            raise

    def _reply_failure(self, reason=None):
        """
        Sends a Failure message to the submitter
        """
        try:
            self.__reply(failure_template, reason)
            log.info('Notified failure of request {} due to: {}'.format(self.request_id, reason))
        except Exception, e:
            # KeyError: Delivery channel data may be invalid
            # IOError: In this case, if even the failure message couldn't be sent, we cannot do anymore :)
            log.warning('Sending failure message for request {}: {}'.format(self.request_id, e.message))

    def submit(self):
        """
        Submit and try to send an acceptance message to the submitter if everything is ok
        """
        try:
            super(DeliveryAction, self).submit()
        except SyntaxError, e:
            # If the message is of bad format, reply notifying the issue
            self._reply_failure(e.message)
            raise
        else:
            try:
                self.__reply_accepted()
            except Exception, e:
                log.warning('Acceptance of request {} failed due to: {}'.format(self.request_id, e.message))
                # If the acceptance message couldn't be sent, remove the request and propagate the error
                self.sink.remove()
                raise

            # If everything was ok, update the request delivery state
            if self.sink.delivery is None:
                self.sink.delivery = 'accepted'


def used_channels():
    """
    Selects all channels that were declared by current requests
    """
    req_channel_keys = r.keys('requests:*:')
    for rck in req_channel_keys:
        try:
            channel = r.hget(rck, 'channel')
            yield channel
        except Exception as e:
            traceback.print_exc()
            log.warning(e.message)


def channel_sharing(channel_b64):
    """
    Calculates how many channel identifiers match the given one (channel_b64)
    """
    return len(list(filter(lambda x: x == channel_b64, used_channels()))) - 1  # Don't count itself


class DeliverySink(Sink):
    __metaclass__ = ABCMeta

    @abstractmethod
    def _save(self, action):
        """
        Stores delivery channel data
        """
        super(DeliverySink, self)._save(action)
        self._pipe.sadd('deliveries', self._request_id)
        broker_b64 = base64.b64encode('|'.join(map(lambda x: str(x), action.request.broker.values())))
        channel_b64 = base64.b64encode('|'.join(action.request.channel.values()))
        self._pipe.hmset('channels:{}'.format(channel_b64), action.request.channel)
        self._pipe.hmset('brokers:{}'.format(broker_b64), action.request.broker)
        self._pipe.hset('{}'.format(self._request_key), 'channel', channel_b64)
        self._pipe.hset('{}'.format(self._request_key), 'broker', broker_b64)

    @abstractmethod
    def _load(self):
        """
        Loads all delivery data
        """
        super(DeliverySink, self)._load()
        self._dict_fields['channel'] = r.hgetall('channels:{}'.format(self._dict_fields['channel']))
        self._dict_fields['broker'] = r.hgetall('brokers:{}'.format(self._dict_fields['broker']))
        self._dict_fields['broker']['port'] = int(self._dict_fields['broker']['port'])
        recipient = self._dict_fields['channel'].copy()
        recipient.update(self._dict_fields['broker'])
        self._dict_fields['recipient'] = recipient

        # If present, remove previously stored delivery state so it can be retrieved each time the delivery getter
        # is invoked
        try:
            del self._dict_fields['delivery']
        except KeyError:
            pass

    @abstractmethod
    def _remove(self, pipe):
        """
        Remove all delivery data
        """

        # If this request is the only one that's using such channel, it is removed
        channel_b64 = r.hget(self._request_key, 'channel')
        sharing = channel_sharing(channel_b64)
        if not sharing:
            log.info('Removing delivery channel ({}) for request {}'.format(channel_b64, self._request_id))
            pipe.delete('channels:{}'.format(channel_b64))
        else:
            log.info('Cannot remove delivery channel of request {}. It is being shared with {} another requests'.format(
                self.request_id, sharing))

        super(DeliverySink, self)._remove(pipe)

        pipe.srem('deliveries', self._request_id)
        pipe.srem('deliveries:ready', self._request_id)

    @property
    def delivery(self):
        return r.hget('{}'.format(self._request_key), 'delivery')

    @delivery.setter
    def delivery(self, value):
        """
        Changes the delivery state of the request
        :param value: 'ready', 'sent', 'accepted', ...
        """
        with r.pipeline(transaction=True) as p:
            p.multi()
            if value == 'ready':
                p.sadd('deliveries:ready', self._request_id)
            elif value == 'sent':
                p.sadd('deliveries:sent', self._request_id)
            if value != 'ready':
                p.srem('deliveries:ready', self._request_id)
            p.hset('{}'.format(self._request_key), 'delivery', value)
            p.execute()
        log.info('Request {} delivery state is now "{}"'.format(self._request_id, value))


class DeliveryResponse(Response):
    __metaclass__ = ABCMeta

    def __init__(self, rid):
        super(DeliveryResponse, self).__init__(rid)

    @abstractmethod
    def build(self):
        super(DeliveryResponse, self).build()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import os
from flask import Flask, jsonify

__author__ = 'Fernando Serena'

config = os.environ.get('STOA_CONFIG', 'agora.stoa.server.config.DevelopmentConfig')

app = Flask(__name__)
app.config.from_object(config)


class APIError(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


class NotFound(APIError):
    def __init__(self, message, payload=None):
        super(NotFound, self).__init__(message, 404, payload)


class Conflict(APIError):
    def __init__(self, message, payload=None):
        super(Conflict, self).__init__(message, 409, payload)


@app.errorhandler(APIError)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import logging
from threading import Thread

import pika
import time

from agora.stoa.actions import execute
from agora.stoa.server import app

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.messaging')

# Load environment variables
BROKER_CONFIG = app.config['BROKER']
EXCHANGE_CONFIG = app.config['EXCHANGE']

exchange = EXCHANGE_CONFIG['exchange']
queue = EXCHANGE_CONFIG['queue']
topic_pattern = EXCHANGE_CONFIG['topic_pattern']
response_prefix = EXCHANGE_CONFIG['response_rk']

log.info("""Broker setup:
                - host: {}
                - port: {}
                - exchange: {}
                - queue: {}
                - topic pattern: {}
                - response prefix: {}""".format(BROKER_CONFIG['host'],
                                                BROKER_CONFIG['port'],
                                                exchange, queue, topic_pattern, response_prefix))


def callback(ch, method, properties, body):
    action_args = method.routing_key.split('.')[2:]
    log.info('--> Incoming {} request!'.format(action_args[0]))
    try:
        execute(*action_args, data=body)
    except (NameError, SystemError) as e:
        log.error(e.message)
        ch.basic_reject(delivery_tag=method.delivery_tag, requeue=True)
        log.debug('Sent REJECT')
    except IOError as e:
        ch.basic_ack(delivery_tag=method.delivery_tag)
        log.error('There was a messaging problem with a request: {}'.format(e.message))
    except Exception as e:
        # EnvironmentError, such as 'Agora is not available' will be caught here...
        ch.basic_ack(delivery_tag=method.delivery_tag)
        log.error(e.message)
    else:
        ch.basic_ack(delivery_tag=method.delivery_tag)
        log.debug('Sent ACK')


def __setup_queues():
    """
    Establish the Stoa messaging system
    """
    while True:

        try:
            connection = pika.BlockingConnection(pika.ConnectionParameters(
                host=BROKER_CONFIG['host']))
        except Exception, e:
            log.error('AMQP broker is not available: {}'.format(e.message))
        else:
            channel = connection.channel()
            log.info('Connected to the AMQP broker: {}'.format(BROKER_CONFIG))

            log.info('Declaring exchange "{}"...'.format(exchange))
            channel.exchange_declare(exchange=exchange,
                                     type='topic', durable=True)

            # Create the requests queue and binding
            channel.queue_declare(queue, durable=True)
            log.info('Declaring queue "{}"...'.format(queue))
            channel.queue_bind(exchange=exchange, queue=queue, routing_key=topic_pattern)
            log.info('Binding to topic "{}"...'.format(topic_pattern))
            channel.basic_qos(prefetch_count=1)
            channel.basic_consume(callback, queue=queue)

            log.info('Ready to accept requests')
            try:
                channel.start_consuming()
            except Exception, e:
                log.error('Messaging system failed due to: {}'.format(e.message))

        time.sleep(1)


# Create and start delivery daemon
th = Thread(target=__setup_queues)
th.daemon = True
th.start()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import calendar
import logging
import shutil

from datetime import datetime as dt

import os
import shortuuid
from agora.stoa.server import app
from agora.stoa.store import r
from agora.stoa.actions.core import STOA
from concurrent.futures import ThreadPoolExecutor
from rdflib import ConjunctiveGraph, URIRef, Literal, XSD, BNode

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.store.triples')
pool = ThreadPoolExecutor(max_workers=4)


def load_stream_triples(fid, until):
    def __triplify(x):
        def __extract_lang(v):
            if '@' in v:
                (v, lang) = tuple(v.split('@'))
            else:
                lang = None
            return v, lang

        def __term(elm):
            if elm.startswith('<'):
                return URIRef(elm.lstrip('<').rstrip('>'))
            elif '^^' in elm:
                (value, ty) = tuple(elm.split('^^'))
                return Literal(value.replace('"', ''), datatype=URIRef(ty.lstrip('<').rstrip('>')))
            elif elm.startswith('_:'):
                return BNode(elm.replace('_:', ''))
            else:
                (elm, lang) = __extract_lang(elm)
                elm = elm.replace('"', '')
                if lang is not None:
                    return Literal(elm, lang=lang)
                else:
                    return Literal(elm, datatype=XSD.string)

        c, s, p, o = eval(x)
        return c, __term(s), __term(p), __term(o)

    for x in r.zrangebyscore('fragments:{}:stream'.format(fid), '-inf', '{}'.format(float(until))):
        yield __triplify(x)


def add_stream_triple(fid, tp, (s, p, o), timestamp=None):
    try:
        if timestamp is None:
            timestamp = calendar.timegm(dt.utcnow().timetuple())
        quad = (tp, s.n3(), p.n3(), o.n3())
        stream_key = 'fragments:{}:stream'.format(fid)
        not_found = not bool(r.zscore(stream_key, quad))
        if not_found:
            with r.pipeline() as pipe:
                pipe.zadd(stream_key, timestamp, quad)
                pipe.execute()
        return not_found
    except Exception, e:
        log.error(e.message)


class GraphProvider(object):
    def __init__(self):
        self.__graph_dict = {}

    @staticmethod
    def __clean(name):
        shutil.rmtree('store/query/{}'.format(name))

    def create(self, conjunctive=False):
        uuid = shortuuid.uuid()
        if conjunctive:
            if 'persist' in app.config['STORE']:
                g = ConjunctiveGraph('Sleepycat')
                g.open('store/query/{}'.format(uuid), create=True)
                g.store.graph_aware = False
            else:
                g = ConjunctiveGraph()
                g.store.graph_aware = False
        else:
            g = query.get_context(uuid)
        self.__graph_dict[g] = uuid
        return g

    def destroy(self, g):
        if g in self.__graph_dict:
            if isinstance(g, ConjunctiveGraph):
                if 'persist' in app.config['STORE']:
                    g.close()
                    pool.submit(self.__clean, self.__graph_dict[g])
                else:
                    g.remove((None, None, None))
                    g.close()
            else:
                query.remove_context(query.get_context(self.__graph_dict[g]))
            del self.__graph_dict[g]


graph_provider = GraphProvider()
store_mode = app.config['STORE']
if 'persist' in store_mode:
    log.info('Creating store folders...')
    if not os.path.exists('store'):
        os.makedirs('store')
    if os.path.exists('store/query'):
        shutil.rmtree('store/query/')
    if os.path.exists('store/tmp'):
        shutil.rmtree('store/tmp')
    os.makedirs('store/query')
    log.info('Loading known triples...')
    cache = ConjunctiveGraph('Sleepycat')
    log.info('Building cache graph...')
    cache.open('store/cache', create=True)
    query = ConjunctiveGraph('Sleepycat')
    log.info('Building temp graph...')
    query.open('store/tmp', create=True)
else:
    cache = ConjunctiveGraph()
    query = ConjunctiveGraph()

cache.store.graph_aware = False
query.store.graph_aware = False
cache.bind('stoa', STOA)
query.bind('stoa', STOA)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

from agora.stoa.server import app, NotFound
from agora.stoa.store import r
from flask import jsonify
from flask.views import View

__author__ = 'Fernando Serena'


def filter_hash_attrs(key, predicate):
    hash_map = r.hgetall(key)
    visible_attrs = filter(predicate, hash_map)
    return {attr: hash_map[attr] for attr in filter(lambda x: x in visible_attrs, hash_map)}


@app.route('/requests')
def get_requests():
    requests = [rk.split(':')[1] for rk in r.keys('requests:*:')]
    return jsonify(requests=requests)


@app.route('/requests/<rid>')
def get_request(rid):
    if not r.exists('requests:{}:'.format(rid)):
        raise NotFound('The request {} does not exist'.format(rid))
    r_dict = filter_hash_attrs('requests:{}:'.format(rid), lambda x: not x.startswith('__'))
    channel = r_dict['channel']
    ch_dict = r.hgetall('channels:{}'.format(channel))
    broker = r_dict['broker']
    br_dict = r.hgetall('brokers:{}'.format(broker))
    r_dict['channel'] = ch_dict
    r_dict['broker'] = br_dict
    if 'mapping' in r_dict:
        r_dict['mapping'] = eval(r_dict['mapping'])

    return jsonify(r_dict)


@app.route('/fragments')
def get_fragments():
    fragment_ids = list(r.smembers('fragments'))
    f_list = [{'id': fid, 'gp': list(r.smembers('fragments:{}:gp'.format(fid)))} for fid in fragment_ids]
    return jsonify(fragments=f_list)


class FragmentView(View):
    decorators = []

    @staticmethod
    def __get_fragment(fid):
        if not r.sismember('fragments', fid):
            raise NotFound('The fragment {} does not exist'.format(fid))

        f_dict = {
            'id': fid,
            'gp': list(r.smembers('fragments:{}:gp'.format(fid))),
            'synced': r.exists('fragments:{}:sync'.format(fid)),
            'requests': list(r.smembers('fragments:{}:requests'.format(fid)))
        }

        return f_dict

    def dispatch_request(self, **kwargs):
        result = self.__get_fragment(kwargs['fid'])
        for decorator in FragmentView.decorators:
            result = decorator(**result)
        return jsonify(result)


app.add_url_rule('/fragments/<fid>', view_func=FragmentView.as_view('fragment'))



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import inspect
import logging

from agora.stoa.actions.core.base import Action
from types import ModuleType

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.actions')
action_modules = {}


def register_module(name, module):
    """
    Registers a Stoa module
    :param name: The unique module name
    :param module: Module instance
    """
    if name in action_modules:
        raise NameError('The module {} already exists'.format(name))
    if not isinstance(module, ModuleType):
        raise AttributeError('{} is not a valid module instance'.format(module))
    action_modules[name] = module


def search_module(module, predicate, limit=1):
    """
    Searches in a module for elements that satisfy a given predicate
    :param module: A module name
    :param predicate: A predicate to check module elements
    :param limit: Default 1
    :return:
    """
    py_mod = action_modules.get(module, None)

    if py_mod is not None:
        cand_elms = filter(predicate,
                           inspect.getmembers(py_mod, lambda x: inspect.isclass(x) and not inspect.isabstract(x)))
        if len(cand_elms) > limit:
            raise EnvironmentError('Too many elements in module {}'.format(module))
        return cand_elms

    return None


def get_instance(module, clz, *args):
    """
    Creates an instance of a given class and module
    :param module: Module name
    :param clz: Instance class
    :param args: Creation arguments
    :return: The instance
    """
    module = action_modules[module]
    class_ = getattr(module, clz)
    instance = class_(*args)
    return instance


def execute(*args, **kwargs):
    """
    Prepares and submits a Stoa-action
    :param args: Action context arguments
    :param kwargs: Action data dictionary
    """

    # The action name (that must match one of the registered modules in order to be submitted)
    name = args[0]
    log.debug('Searching for a compliant "{}" action handler...'.format(name))

    try:
        _, clz = search_module(name,
                               lambda (_, cl): issubclass(cl, Action) and cl != Action).pop()

    except EnvironmentError:
        raise SystemError('Cannot handle {} requests'.format(name))
    except IndexError:
        raise ("Couldn't find an Action class inside {} module".format(name))

    try:
        # Extract the request message from kwargs,
        data = kwargs.get('data', None)
        log.debug(
            'Found! Requesting an instance of {} to perform a/n {} action described as:\n{}'.format(clz, name,
                                                                                                    data))

        # Create the proper action instance...
        action = clz(data)
    except IndexError:
        raise NameError('Action module found but class is missing: "{}"'.format(name))
    else:
        # and submit!
        rid = action.submit()
        if rid is not None:
            log.info('A {} request was successfully submitted with id {}'.format(name, rid))



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import traceback
from threading import Thread

from agora.stoa.messaging.reply import reply
from agora.stoa.server import app
from agora.stoa.store import r
from concurrent.futures.thread import ThreadPoolExecutor

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.daemons.delivery')

# Load environment variables
MAX_CONCURRENT_DELIVERIES = int(app.config.get('PARAMS', {}).get('max_concurrent_deliveries', 8))

# Delivery thread pool
thp = ThreadPoolExecutor(max_workers=min(8, MAX_CONCURRENT_DELIVERIES))

log.info("""Delivery daemon setup:
                    - Maximum concurrent deliveries: {}""".format(MAX_CONCURRENT_DELIVERIES))


def build_response(rid):
    """
    Creates a response instance for a given request id
    :param rid: Request identifier
    :return: The response object
    """
    from agora.stoa.actions import get_instance
    response_class = r.hget('requests:{}:'.format(rid), '__response_class')
    if response_class is None:
        raise AttributeError('Cannot create a response for {}'.format(rid))
    (module_name, class_name) = tuple(response_class.split('.'))
    return get_instance(module_name, class_name, rid)


def __deliver_response(rid):
    """
    The delivery task for a given request id
    :param rid: Request id
    """

    def deliver_message():
        reply(message, headers=headers, **response.sink.recipient)
        return len(str(message))

    response = None
    try:
        response = build_response(rid)
        delivery_state = response.sink.delivery
        if delivery_state == 'ready':
            messages = response.build()
            # The creation of a response object may change the corresponding request delivery state
            # (mixing, streaming, etc). The thing is that it was 'ready' before,
            # so it should has something prepared to deliver.
            n_messages = 0
            deliver_weight = 0

            message, headers = messages.next()  # Actually, this is the trigger
            deliver_weight += deliver_message()
            n_messages += 1

            deliver_delta = 0
            for (message, headers) in messages:
                message_weight = deliver_message()
                deliver_weight += message_weight
                deliver_delta += message_weight
                n_messages += 1
                if deliver_delta > 1000:
                    deliver_delta = 0
                    log.info('Delivering response of request {} [{} kB]'.format(rid, deliver_weight / 1000.0))

            deliver_weight /= 1000.0
            log.info('{} messages delivered for request {} [{} kB]'.format(n_messages, rid, deliver_weight))

        elif delivery_state == 'accepted':
            log.error('Request {} should not be marked as deliver-ready, its state is inconsistent'.format(rid))
        else:
            log.info('Response of request {} is being delivered by other means...'.format(rid))
            r.srem('deliveries:ready', rid)
    except StopIteration:  # There was nothing prepared to deliver (Its state may have changed to
        # 'streaming')
        r.srem('deliveries:ready', rid)
    except (EnvironmentError, AttributeError, Exception), e:
        r.srem('deliveries:ready', rid)
        traceback.print_exc()
        log.warning(e.message)
        if response is not None:
            log.error('Force remove of request {} due to a delivery error'.format(rid))
            response.sink.remove()
        else:
            log.error("Couldn't remove request {}".format(rid))


def __deliver_responses():
    import time
    log.info('Delivery daemon started')

    # Declare in-progress deliveries dictionary
    futures = {}
    while True:
        # Get all ready deliveries
        ready = r.smembers('deliveries:ready')
        for rid in ready:
            # If the delivery is not in the thread pool, just submit it
            if rid not in futures:
                log.info('Response delivery of request {} is ready. Putting it in queue...'.format(rid))
                futures[rid] = thp.submit(__deliver_response, rid)

        # Clear futures that have already ceased to be ready
        for obsolete_rid in set.difference(set(futures.keys()), ready):
            if obsolete_rid in futures and futures[obsolete_rid].done():
                del futures[obsolete_rid]

        # All those deliveries that are marked as 'sent' are being cleared here along its request data
        sent = r.smembers('deliveries:sent')
        for rid in sent:
            r.srem('deliveries:ready', rid)
            r.srem('deliveries', rid)
            try:
                response = build_response(rid)
                response.sink.remove()
                log.info('Request {} was sent and cleared'.format(rid))
            except AttributeError:
                log.warning('Request number {} was deleted by other means'.format(rid))
                pass
            r.srem('deliveries:sent', rid)

        time.sleep(0.1)


# Log delivery counters at startup
registered_deliveries = r.scard('deliveries')
deliveries_ready = r.scard('deliveries:ready')
log.info("""Delivery daemon started:
                - Deliveries: {}
                - Ready: {}""".format(registered_deliveries, deliveries_ready))

# Create and start delivery daemon
__thread = Thread(target=__deliver_responses)
__thread.daemon = True
__thread.start()
